#!/usr/bin/env python
###############################################################################
##
## MODULE      : tikz.py
## DESCRIPTION : TikZ support
## COPYRIGHT   : (C) 2019  Darcy Shen, Massimiliano Gubinelli
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .latex import LaTeX
from ..protocol import *


class TikZ(LaTeX):
    def __init__(self, name = "tikz"):
        super(TikZ, self).__init__()
        self.name = name

        self.pre_code = """
\\documentclass[tikz]{standalone}
\\begin{document}
"""
        self.post_code = """
\\end{document}
"""
        self.message = "TeXmacs interface to TikZ"

    def available(self):
        if not super(TikZ, self).available():
            return False
        for sty in ("standalone", "tikz"):
            if len (super(TikZ, self).kpsewhich(sty + ".sty")) <= 0:
                flush_err ("Failed to find " + sty +".sty,"
                           " please install the missing LaTeX packages\n")
                return False
        return True
        
    def evaluate(self, code):
        if not (code.lstrip().startswith("\\documentclass")):
            if code.lstrip().startswith("\\usetikzlibrary"):
                pass
            elif not (code.lstrip().startswith("\\begin{tikzpicture}")):
                code = "\\begin{tikzpicture}\n" + code + "\n\\end{tikzpicture}"
            code = self.pre_code + "\n" + code + "\n" + self.post_code

        super(TikZ, self).evaluate(code)



#!/usr/bin/env python
###############################################################################
#
# MODULE      : completion.py
# DESCRIPTION : Initialize python plugin
# COPYRIGHT   : (C) 2004  Ero Carrera, ero@dkbza.org
#               (C) 2019  Darcy Shen
#
# This software falls under the GNU general public license version 3 or later.
# It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
# in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import sys
import os
import keyword

py_ver = sys.version_info[0]

def do_module_hierarchy(mod, attr):
    """Explore an object's hierarchy.
    
    Go through the object hierarchy looking for
    attributes/methods to provide as autocompletion options.
    """
    dot = attr.find('.')
    if dot>0:
       if hasattr(mod, attr[:dot]):
          next = getattr(mod, attr[:dot])
          return do_module_hierarchy(next, attr[dot+1:])
    if isinstance(mod, dict):
       return dir(mod)
    else:
       return dir(mod)
 

def find_completion_candidates(cmpl_str, my_globals):
    """Harvest candidates to provide as autocompletion options."""
    
    if py_ver == 3:
        haystack = list(my_globals.keys()) + \
                   dir(my_globals['__builtins__']) + keyword.kwlist
    else:
        haystack = my_globals.keys() + \
                   dir(my_globals['__builtins__']) + keyword.kwlist

    dot = cmpl_str.rfind('.')
    offset = None
    if dot >= 0:
        offset = len(cmpl_str[dot+1:])
        first_dot = cmpl_str[:dot].find('.')
        if first_dot < 0:
            mod_name = cmpl_str[:dot]
            r_str = cmpl_str[dot+1:]
        else:
            mod_name = cmpl_str[:first_dot]
            r_str = cmpl_str[first_dot+1:]
        if mod_name in keyword.kwlist:
            return None, []
        if py_ver == 3:    
            if mod_name in os.sys.modules:
                haystack = do_module_hierarchy(os.sys.modules[mod_name], r_str)
            elif mod_name in list(my_globals.keys()):
                haystack = do_module_hierarchy(my_globals[mod_name], r_str)
            else:
                haystack = do_module_hierarchy(type(mod_name), r_str)
        else:
            if os.sys.modules.has_key(mod_name):
                haystack = do_module_hierarchy(os.sys.modules[mod_name], r_str)
            elif mod_name in my_globals.keys():
                haystack = do_module_hierarchy(my_globals[mod_name], r_str)
            else:
                haystack = do_module_hierarchy(type(mod_name), r_str)
            
    if py_ver == 3:
        return offset, [x for x in haystack if x.find(cmpl_str[dot+1:])  ==  0]
    else:
        return offset, filter(lambda x:x.find(cmpl_str[dot+1:])  ==  0, haystack)


def name_char(c):
    """Check whether a character is a valid identifier/keyword."""
    return c not in '+-*/%<>&|^~=!,:()[]{} \n\t'


def complete(s, pos, my_globals):
    """Process autocomplete command. """
    
    try:
        s = s[:pos]
        if not s:
            return '(tuple "" "")'
    except Exception as e:
        return '(tuple "" "")'
    # We get the string after the last space character.
    # No completion is done for strings containing spaces.
    i = len(s) - 1
    while i > 0:
        if not name_char(s[i]):
            i += 1
            break
        i -= 1
    s = s[i:]
    pos = len(s)
    # no string after last space? return empty completion
    if not s:
        return '(tuple "" "")'
        
    # Find completion candidates and form a suitable answer to Texmacs
    offset, cand = find_completion_candidates (s, my_globals)
    if not cand:
        res = '""'
    else:
        res = ''
    for c in cand:
        if offset is not None:
            pos = offset
        res += '"%s" ' % c[pos:]
    return '(tuple "' + s + '" ' + res + ')'

def from_scm_string(s):
    if len(s) > 2 and s[0] == '"' and s[-1] == '"':
        return s[1:-1]
    return s


def parse_complete_command(s):
    """HACK"""
    t1 = s.strip().strip('()').split(' ', 1)
    t2 = t1[1].rsplit(' ', 1)
    # Don't use strip('"') in case there are several double quotes
    return [t1[0], from_scm_string(t2[0]), int(t2[1])]



#!/usr/bin/env python
###############################################################################
#
# MODULE      : compat.py
# DESCRIPTION : Code from python stdlib for compatibility
#               Should be removed when we upgrade to Python 3

import os
import sys

py_ver = sys.version_info[0]

def tm_input():
    if py_ver == 3:
        return input()
    else:
        return raw_input()

# Check that a given file can be accessed with the correct mode.
# Additionally check that `file` is not a directory, as on Windows
# directories pass the os.access check.
def _access_check(fn, mode):
    return (os.path.exists(fn) and os.access(fn, mode)
            and not os.path.isdir(fn))

encoding = sys.getfilesystemencoding()

def fsencode(filename):
    if isinstance(filename, str):
        return filename.encode(encoding)
    else:
        return filename

def fsdecode(filename):
    if isinstance(filename, bytes):
        return filename.decode(encoding)
    else:
        return filename

# Extracted from cpython: Lib/shutil.py
def which(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.
    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.
    """
    # If we're given a path with a directory part, look it up directly rather
    # than referring to PATH directories. This includes checking relative to the
    # current directory, e.g. ./script
    if os.path.dirname(cmd):
        if _access_check(cmd, mode):
            return cmd
        return None

    use_bytes = isinstance(cmd, bytes)

    if path is None:
        path = os.environ.get("PATH", None)
        if path is None:
            try:
                path = os.confstr("CS_PATH")
            except (AttributeError, ValueError):
                # os.confstr() or CS_PATH is not available
                path = os.defpath
        # bpo-35755: Don't use os.defpath if the PATH environment variable is
        # set to an empty string

    # PATH='' doesn't match, whereas PATH=':' looks in the current directory
    if not path:
        return None

    if use_bytes:
        path = fsencode(path)
        path = path.split(fsencode(os.pathsep))
    else:
        path = fsdecode(path)
        path = path.split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        curdir = os.curdir
        if use_bytes:
            curdir = fsencode(curdir)
        if curdir not in path:
            path.insert(0, curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        if use_bytes:
            pathext = [fsencode(ext) for ext in pathext]
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        # If it does match, only test that one, otherwise we have to try
        # others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir in path:
        normdir = os.path.normcase(dir)
        if not normdir in seen:
            seen.add(normdir)
            for thefile in files:
                name = os.path.join(dir, thefile)
                if _access_check(name, mode):
                    return name
    return None


#!/usr/bin/env python3
###############################################################################
##
## MODULE      : tm_eukleides.py
## DESCRIPTION : Launcher for the eukleides plugin
## COPYRIGHT   : (C) 2021  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists

tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")

from subprocess import Popen, PIPE, STDOUT
from tmpy.graph.graph import Graph
from tmpy.protocol import *
from tmpy.capture import CaptureStdout
from tmpy.compat import *


class Eukleides(Graph):
    def __init__(self, name = "eukleides"):
        super(Eukleides, self).__init__()
        self.name = name
        self.width = 600
        self.default_output = "eps"

    def greet(self):
        flush_latex ("E \\Upsilon K \\Lambda \\tmop{EI} \\Delta H \\Sigma")
        flush_newline ()
        flush_verbatim ("A Euclidean Geometry Drawing Language")
        flush_newline ()
        try:
            p = Popen([self.name, "-v"], stdout=PIPE)
            ret, err = p.communicate()
            # WARN: The Version Info is in stderr
            if (p.returncode == 0):
                self.message = ret.decode()
        except OSError:
            pass
        super(Eukleides, self).greet()

    def evaluate(self, code):
        if self.output == "eps":
            image = self.get_eps()
            image_path = self.get_eps_path()
        elif self.output == "png":
            image = self.get_png()
        elif self.output == "svg":
            image = self.get_svg()
        else:
            flush_verbatim("Unsupported output type: " + self.output)
            return

        code_path = self.get_tmp_dir() + self.name + ".txt"
        with open(code_path, 'w') as code_file:
            code_file.write(self.pre_code)
            code_file.write(code)

        cmd = [self.name, f"--output={image_path}", code_path]
        p = Popen(cmd, stderr=PIPE)
        out, err = p.communicate()
        if (p.returncode == 0):
          flush_file (image)
        else:
          flush_verbatim (err.decode())

    def main_loop(self):
        # Main session loop.
        while True:
            line = tm_input()
            if not line:
                continue
            if line[0] == DATA_COMMAND:
                # TODO: Handle completions
                continue
            else:
                lines = []
                for x in line.split('~'):
                    lines.append(x)
                while line != "<EOF>":
                    line = tm_input()
                    for x in line.split('~'):
                        lines.append(x)
                text = '\n'.join(lines[:-1])
                self.eval(text)


if (exists (tmpy_home_path)):
        flush_verbatim ("WARNING: You are under develop mode using " + tmpy_home_path)
        flush_newline (2)

my_globals   = {}

text = 'import builtins as __builtins__'
CaptureStdout.capture (text, my_globals, "tm_eukleides")

current = Eukleides()
if not current.available():
    flush_err ("Failed to launch the Eukleides plugin, aborted!!!")
    exit(-1)

current.greet()
current.main_loop()



#!/usr/bin/env python
###############################################################################
##
## MODULE      : graphviz.py
## DESCRIPTION : Graphviz plotting support
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import sys
from subprocess import Popen, PIPE, STDOUT
from .graph import Graph
from ..protocol import *

class Graphviz(Graph):
    def __init__(self, name):
        super(Graphviz, self).__init__()
        self.name = name
        self.default_output = "eps"
    
    def greet(self):
        if len(self.message) == 0:
            try:
                p = Popen([self.name, "-V"], stderr=PIPE)
                ret, err = p.communicate()
                # WARN: The Version Info is in stderr
                if (p.returncode == 0):
                    self.message = err.decode()
            except OSError:
                pass
        super(Graphviz, self).greet()

    def evaluate(self, code):
        # NOTE: the eps output format does not always work
        # for example, when we use Chinese in a label
        # by default, we use eps, however, we may switch to png via
        # the -output option on the magic line
        if self.output == "eps":
            path = self.get_eps_path()
            image = self.get_eps()
            cmd_list = [self.name, "-Teps"]
        elif self.output == "png":
            path = self.get_png_path()
            image = self.get_png()
            cmd_list = [self.name, "-Tpng"]
        elif self.output == "svg":
            path = self.get_svg_path()
            image = self.get_svg()
            cmd_list = [self.name, "-Tsvg"]
        else:
            flush_verbatim("Unsupported output type: " + self.output)
            return

        if os.path.isfile(path):
            os.remove(path)
        f = open(path, 'wb')
        p = Popen(cmd_list, stdout=f, stdin=PIPE, stderr=PIPE)
        py_ver = sys.version_info[0]
        if py_ver == 3:
            out, err = p.communicate(input=code.encode())
        else:
            out, err = p.communicate(input=code)
        if (p.returncode == 0):
            flush_file (image)
        else:
            flush_verbatim (err.decode())



#!/usr/bin/env python
###############################################################################
##
## MODULE      : protocol.py
## DESCRIPTION : The TeXmacs plugin protocol impl
## COPYRIGHT   : (C) 2004  Ero Carrera, ero@dkbza.org
##               (C) 2012  Adrian Soto
##               (C) 2014  Miguel de Benito Delgado, mdbenito@texmacs.org
##               (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from os.path import exists

DATA_BEGIN = chr(2)
DATA_END = chr(5)
DATA_ESCAPE = chr(27)
DATA_COMMAND = chr(16)


def data_begin():
    """Signal the beginning of data to TeXmacs."""
    os.sys.stdout.write(chr(2))


def data_end():
    """Signal the end of data to TeXmacs."""
    os.sys.stdout.write(chr(5))
    os.sys.stdout.flush()


def texmacs_escape (data):
    return data.replace(DATA_BEGIN.encode(), (DATA_ESCAPE + DATA_BEGIN).encode()) \
               .replace(DATA_END.encode(), (DATA_ESCAPE + DATA_END).encode())


def flush_any (out_str):
    """Feed data back to TeXmacs.

    Output results back to TeXmacs, with the DATA_BEGIN,
    DATA_END control characters."""
    data_begin()
    os.sys.stdout.write(out_str)
    data_end()

def flush_verbatim(content):
    flush_any ("utf8:" + content)

def flush_newline (n = 1):
    if (n >= 1):
        for i in range(n):
            flush_verbatim("\n")

def flush_prompt(prompt):
    flush_any ("prompt#" + prompt)

def flush_command(command):
    flush_any ("command:" + command)

def flush_scheme(scheme):
    flush_any ("scheme:" + scheme)

def flush_latex(latex):
    flush_any ("latex:$" + latex + "$")

def flush_file(path):
    flush_any ("file:" + path)

def flush_ps(content):
    flush_any ("ps:" + content)

def flush_err(content):
    os.sys.stderr.write(chr(2))
    os.sys.stderr.write("utf8:" + content)
    os.sys.stderr.write(chr(5))
    os.sys.stderr.flush() 

def get_plugin_path(name):
    the_home_path = os.getenv("TEXMACS_HOME_PATH").replace("\\", "/") + "/plugins/" + name
    the_sys_path = os.getenv("TEXMACS_PATH").replace("\\", "/") + "/plugins/" + name
    if (exists(the_home_path)):
        return the_home_path
    else:
        return the_sys_path



import os
import traceback

class CaptureStdout:
    """Capture output to os.sys.stdout.

    Class in charge of recording the output of the
    statements/expressions entered in the TeXmacs
    session and executed in Python.

    Must be used in a with statement, as in CaptureStdout.capture()
    """

    def __enter__(self):
        """ """
        class Capture:
            def __init__(self):
                self.text = ''
            def write(self, str):
                self.text += str
            def flush(self):
                os.sys.stdout.flush() # Needed?
                self.text = ''
            def getOutput(self):
                return self.text

        self.capt = Capture()
        self.stdout_saved, os.sys.stdout = os.sys.stdout, self.capt        
        return self.capt
    
    def __exit__(self, type, value, traceback):
        os.sys.stdout = self.stdout_saved

    @staticmethod
    def capture (code, env, filename):
        with CaptureStdout() as capt:
            try:
                eval (compile (code, filename, 'exec'), env)
            except Exception as e:
                traceback.print_exc (file = os.sys.stdout, limit = 0)
            return capt.getOutput()




#!/usr/bin/env python
###############################################################################
##
## MODULE      : tm_graphviz.py
## DESCRIPTION : Launcher for the Graphviz plugin
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists
tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")

from tmpy.graph.graphviz  import Graphviz
from tmpy.capture         import CaptureStdout
from tmpy.compat          import py_ver
from tmpy.protocol        import *
from tmpy.compat          import *

if (exists (tmpy_home_path)):
    flush_verbatim ("WARNING: You are under develop mode using " + tmpy_home_path)
    flush_newline (2)

my_globals   = {}

if py_ver == 3:
    text = 'import builtins as __builtins__'
else:
    text = 'import __builtin__ as __builtins__'
CaptureStdout.capture (text, my_globals, "tm_graphviz")

graphs = list(map(lambda x: Graphviz(x), ["dot", "neato", "twopi", "circo", "fdp", "sfdp", "patchwork", "osage"]))
graph_names = list(map(lambda x: x.name, graphs))

if len(graphs) == 0:
    flush_verbatim ("\nSorry, please check your installation of Graphviz")
    flush_prompt ("dead] ")
    exit(0)

current = graphs[0]
current.greet()
flush_prompt (current.name + "] ")

def unigraph(text):
    magic_lines = text.split("\n")
    magic_line = magic_lines[0]
    command = magic_line.split(" ")[0].strip("%")

    if command in graph_names:
        graph = graphs[graph_names.index(command)]
        graph.eval(text)
    else:
        flush_verbatim ("No such Graphs backend: " + command)


# Main session loop.
while True:
    line = tm_input()
    if not line:
        continue
    if line[0] == DATA_COMMAND:
        # TODO: Handle completions
        continue
    else:
        lines = [line]
        while line != "<EOF>":
            line = tm_input()
            if line == '': 
                continue
            lines.append(line)

        text='\n'.join(lines[:-1])

        if text.startswith("%"):
            unigraph(text)
        elif text == "help":
            flush_verbatim ("[help, " + ", ".join(str(x) for x in graph_names) + "]\n")
            current.greet()
        elif text in graph_names:
            current = graphs[graph_names.index(text)]
            current.greet()
        else:
            current.eval(text)



#!/usr/bin/env python
###############################################################################
##
## MODULE      : tm_sympy.py
## DESCRIPTION : Launcher for the SymPy plugin
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists
tmpy_home_path= os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")


from tmpy.protocol import *
from tmpy.compat import *
from tmpy.completion import parse_complete_command, complete
from tmpy.postscript import *
from tmpy.capture import CaptureStdout

try:
    from sympy.printing import latex
    from sympy import Basic, MatrixBase
    import sympy
except ModuleNotFoundError:
    flush_verbatim ("Please install sympy first, for example:\n")
    flush_verbatim ("    pip install sympy")
    flush_prompt ("dead] ")
    sys.exit(0)

import platform

my_globals   = {}

# We insert into the session's namespace the 'ps_out' method.
my_globals['ps_out'] = ps_out

# As well as some documentation.
my_globals['__doc__'] = """A SymPy plugin for TeXmacs."""

def flush_output (data):
    if isinstance (data, Basic) or isinstance (data, MatrixBase):
        flush_latex (latex (data))
    else:
        flush_verbatim (str(data).strip())


###############################################################################
# Session start
###############################################################################
if (exists (tmpy_home_path)):
    flush_verbatim ("WARNING: You are under develop mode using " + tmpy_home_path)
    flush_newline (2)
flush_verbatim ("SymPy %s under Python %s\n" % (sympy.__version__, platform.python_version()))
flush_verbatim ("Please see the documentation in Help -> Plugins -> SymPy\n")
flush_prompt (">>> ")

while True:
    line= tm_input ()
    if not line:
        continue
    if line[0] == DATA_COMMAND:
        sf = parse_complete_command (line[1:])
        if sf[0] == 'complete':
            flush_scheme (complete (sf[1], sf[2], my_globals))
        continue

    lines = [line]
    while line != "<EOF>":
        line = tm_input()
        if line == '': 
            continue
        lines.append(line)
    text = '\n'.join(lines[:-1])
    try: # Is it an expression?
        result = eval(text, my_globals)
    except:
        result = CaptureStdout.capture(text, my_globals, "tm_sympy")
    flush_output (result)



#!/usr/bin/env python
###############################################################################
##
## MODULE      : pdflatex.py
## DESCRIPTION : LaTeX graphics support
## COPYRIGHT   : (C) 2019  Darcy Shen and Massimiliano Gubinelli
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .graph import Graph
from ..compat import which
from ..protocol import *


class PDFLaTeX(Graph):
    def __init__(self, name = "pdflatex"):
        super(PDFLaTeX, self).__init__()
        self.name = name
        self.message = "TeXmacs interface to PDFLaTeX for picture generation"

    def available(self):
        return all(list(map(lambda x: which(x) is not None, ["pdflatex"])))

    def evaluate(self, code):

        if not (code.lstrip().startswith("\\documentclass")):
            flush_err("I expect a valid LaTeX document, usually starting with\n"
                +"   \\documentclass{standalone}\n"
                +"which will take proper care of suitable bounding box for the image.\n"
                +"The document is sent to latex and then to dvips to produce a single image.")    
            flush_verbatim("")
            return

        code_path = self.get_tmp_dir() + self.name + ".tex"
        pdf_path = self.get_tmp_dir() + self.name + ".pdf"

        with open(code_path, 'w') as code_file:
            code_file.write(code)

        cmd0 = ["pdflatex", "--interaction=errorstopmode", "-halt-on-error", code_path]

        # flush_err (code) # to debug
        phase = "PDFLaTeX"
        #flush_err ("Running " + phase)
        os.chdir(self.get_tmp_dir())
        p = Popen(cmd0, stdout=PIPE, stderr=STDOUT)
        out, err = p.communicate()
        if (p.returncode == 0):
            phase = "Finished"
            #flush_err ("Running " + phase)
            flush_file (pdf_path)
            return
        flush_err (phase + " error!\n" + out.decode())
        flush_verbatim("")   # otherwise TeXmacs keeps waiting output
            



#!/usr/bin/env python
###############################################################################
##
## MODULE      : dratex.py
## DESCRIPTION : DraTeX support
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .latex import LaTeX
from ..protocol import *


class DraTeX(LaTeX):
    def __init__(self, name = "dratex"):
        super(DraTeX, self).__init__()
        self.name = name
        the_plugin_path = get_plugin_path(name)
        self.pre_code = """
\\documentclass{standalone}
\\input "%s/latex/DraTex.sty"
\\input "%s/latex/AlDraTex.sty"
\\pagestyle{empty}
\\begin{document}
""" % (the_plugin_path, the_plugin_path)
        self.post_code = """
\\end{document}
"""
        self.message = "TeXmacs interface to (Al)DraTex (High Level Drawing Facilities)"

    def evaluate(self, code):
        if not (code.lstrip().startswith("\\documentclass")):
            code = self.pre_code + "\n" + code + "\n" + self.post_code

        super(DraTeX, self).evaluate(code)




#!/usr/bin/env python
###############################################################################
##
## MODULE      : graph.py
## DESCRIPTION : An abstract Graph class
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import platform
import shutil
import time
from ..protocol import *
from ..compat import *


class Graph(object):
    name = ""
    message = ""
    pre_code = ""
    post_code = ""
    height = ""
    width = ""
    output = ""
    default_width = "0px"
    default_height = "0px"
    default_output = "eps"

    def greet(self):
        for x in self.message.split("\n"):
            if len(x) == 0:
                pass
            else:
                flush_verbatim (x + "\n")
                flush_prompt (self.name + "] ")

    def available(self):
        return which(self.name) is not None

    def reset_options(self):
        self.height = self.default_height
        self.width = self.default_width
        self.output = self.default_output

    def apply_magic(self, magic_line):
        args = list(filter(lambda x: len(x)!=0, magic_line.split(" ")[1:]))
        while (len(args) > 1):
            option = args[0]
            value = args[1]
            if (option == '-width'):
                self.width = value
            elif (option == '-height'):
                self.height = value
            elif (option == '-output'):
                self.output = value
            else:
                pass
            args = args[2:]

    def before_evaluate(self):
        self.reset_options()
        if not os.path.exists(self.get_tmp_dir()):
            os.mkdir(self.get_tmp_dir())

    def evaluate(self, code):
        pass

    def after_evaluate(self):
        self.clean_tmp_dir()
        self.remove_tmp_dir()

    def eval(self, code):
        self.before_evaluate()
        if (code.startswith("%")):
            magic_lines = code.split("\n")
            magic_line = magic_lines[0]
            self.apply_magic(magic_line)
            code = '\n'.join(magic_lines[1:])
        self.evaluate(code)
        time.sleep(1)
        self.after_evaluate()

    def main_loop(self):
        # Main session loop.
        while True:
            line = tm_input()
            if not line:
                continue
            if line[0] == DATA_COMMAND:
                # TODO: Handle completions
                continue
            else:
                lines = [line]
                while line != "<EOF>":
                    line = tm_input()
                    lines.append(line)
                text = '\n'.join(lines[:-1])
                self.eval(text)

    def get_tmp_dir(self):
        dir = "graph_" + self.name + "_" + str(os.getpid())
        if (platform.system() == "Windows"):
            return os.getenv("TEXMACS_HOME_PATH") + "\\system\\tmp\\" + dir + "\\"
        else:
            return os.getenv("TEXMACS_HOME_PATH") + "/system/tmp/" + dir + "/"

    def remove_tmp_dir(self):
        if (platform.system() != "Windows"):
            os.rmdir(self.get_tmp_dir())

    def clean_tmp_dir(self):
        folder = self.get_tmp_dir()
        for the_file in os.listdir(folder):
            file_path = os.path.join(folder, the_file)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                print(e)

    def get_png_path(self):
        return self.get_tmp_dir() + self.name + ".png"

    def get_eps_path(self):
        return self.get_tmp_dir() + self.name + ".eps"

    def get_svg_path(self):
        return self.get_tmp_dir() + self.name + ".svg"

    def get_png(self):
        return self.get_png_path() +\
            "?" + "width=" + str(self.width) +\
            "&" + "height=" + str(self.height)

    def get_eps(self):
        return self.get_eps_path() +\
            "?" + "width=" + str(self.width) +\
            "&" + "height=" + str(self.height)

    def get_svg(self):
        return self.get_svg_path() +\
            "?" + "width=" + str(self.width) +\
            "&" + "height=" + str(self.height)



#!/usr/bin/env python
###############################################################################
#
# MODULE      : tm_sage.py
# DESCRIPTION : Initialize python plugin
# COPYRIGHT   : (C) 2004  Ero Carrera, ero@dkbza.org
#               (C) 2019  Darcy Shen
#
# This software falls under the GNU general public license version 3 or later.
# It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
# in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
from os.path import exists
tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")


import tempfile
import traceback
import re
import string
import ast
import warnings
warnings.simplefilter("ignore") # don't print warnings to stdout
from tmpy.compat import py_ver, tm_input
from tmpy.capture import CaptureStdout
from tmpy.postscript import ps_out, PSOutDummy, pdf_out, FileOutDummy
from tmpy.completion import parse_complete_command, complete
from tmpy.protocol   import *
from sage.all        import *

if py_ver == 2:
    flush_err ("This version of SageMath uses Python 2, which is no longer actively supported in TeXmacs.\nIf you experience any problems please upgrade to SageMath 9.0 or higher.")
#    exit (-1)

# import logging as log
# log.basicConfig(filename='/tmp/tm_python.log',level=log.INFO)

def flush_output (data):
# TODO: integrate Sage to TeXmacs parsing from compose_output()
    """Do some parsing on the output according to its type.
    
    Non printable characters in unicode strings are escaped
    and objects of type None are not printed (so that procedure calls,
    as opposed to function calls, don't produce any output)."""

    if (data is None):
        flush_verbatim ("")
        return

    #If the object is a graphics object, try to return
    #postscript to TeXmacs
    if isinstance(data, Graphics):
        try:
            #Save it to a file
            filename = tempfile.mktemp(suffix='.ps')
            data.save(filename)

            #Read the file
            ps_file = open(filename)
            ps_contents = ps_file.read()
            ps_file.close()

            flush_ps(ps_contents)
        except:
            pass
    elif isinstance(data, SageObject):
        try:
            #Replace latex arrays with matrix
            flush_latex(latex(data))
        except:
            flush_verbatim('%s' % str(data))
    elif isinstance(data, str):
        flush_verbatim(data.strip())
    elif isinstance(data, int):
        flush_verbatim('%d' % data)
    elif isinstance(data, float):
        flush_verbatim('%f' % data)
    elif isinstance (data, PSOutDummy):
        flush_ps (data.content)
    elif isinstance (data, FileOutDummy):
        if (data.content is None):
            flush_verbatim ("")
        else:
            flush_file (data.content)
    else:
        flush_verbatim (str(data).strip())

def my_eval (code, p_globals):
    '''Execute a script and return the value of the last expression'''

    block = ast.parse(code, mode='exec')
    if len(block.body) > 1 and isinstance(block.body[-1], ast.Expr):
        last = ast.Expression(block.body.pop().value)
        exec(compile(block, '<string>', mode='exec'), p_globals)
        return eval(compile(last, '<string>', mode='eval'), p_globals)
    else:
        return eval(code, p_globals)

__version__ = '0.8.1'
__author__ = 'Ero Carrera'
my_globals   = {}

# We insert into the session's namespace the 'ps_out' method.
my_globals['ps_out'] = ps_out
my_globals['pdf_out'] = pdf_out

# As well as some documentation.
my_globals['__doc__'] = """A SageMath plugin for TeXmacs.

  TeXmacs SAGE interface v0.8.1.

  Based on the TeXmacs Python plugin by Ero Carrera (c) 2004
  
  The version distributed with TeXmacs is always the latest.

  Enjoy it!
  """

if py_ver > 3:
    text = 'import builtins as __builtins__'
else:
    text = 'import __builtin__ as __builtins__'
CaptureStdout.capture (text, my_globals, "tm_sage")

sys.stdout = os.fdopen (sys.stdout.fileno(), 'w')

co = compile('from sage.all import *', 'tm_sage', 'exec')
eval(co, my_globals)
co = compile('from sage.calculus.predefined import x', 'tm_sage', 'exec')
eval(co, my_globals)

###############################################################################
# Session start
###############################################################################
if (os.path.exists (tmpy_home_path)):
    flush_verbatim ("WARNING: You are under develop mode using " + tmpy_home_path)
    flush_newline (2)
flush_verbatim (sage.misc.banner.version() + "\n" +
		"Python " + sys.version + "\n" +
               "SageMath plugin for TeXmacs.\n" +
               "Please see the documentation in Help -> Plugins -> Sage")
flush_prompt (">>> ")
while True:
    line= tm_input ()
    if not line:
        continue
    if line[0] == DATA_COMMAND:
        sf = parse_complete_command (line[1:])
        if sf[0] == 'complete':
            flush_scheme (complete (sf[1], sf[2], my_globals))
        continue
    elif line.endswith('??') and not line.strip().startswith('#'):
        result = eval('sage.misc.sageinspect.sage_getsource('+line[:-2]+')', my_globals)
    elif line.endswith('?') and not line.strip().startswith('#'):
        if len(line) > 1:
            result = eval('sage.misc.sageinspect.sage_getdoc('+line[:-1]+')', my_globals)
        else:
            flush_verbatim ('Type a name before the "?" to see the help')
        continue
    else:
        lines= [line]
        while line != "<EOF>":
            line= tm_input ()
            if line == '': 
                continue
            lines.append (line)
        text = '\n'.join (lines[:-1])
        try: # Is it an expression?
            result= my_eval (preparse(text), my_globals)
        except:
            result= CaptureStdout.capture (preparse(text), my_globals, "tm_sage")
        flush_output (result)



#!/usr/bin/env python
###############################################################################
##
## MODULE      : mermaid.py
## DESCRIPTION : Mermaid plotting support
## COPYRIGHT   : (C) 2019  Darcy Shen
##
## This software falls under the GNU general public license version 3 or later.
## It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
## in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
from subprocess import Popen, PIPE, STDOUT
from .graph import Graph
from ..protocol import *

class Mermaid(Graph):
    def __init__(self, name = "mmdc"):
        super(Mermaid, self).__init__()
        self.name = name

    def greet(self):
        if len(self.message) == 0:
            try:
                p = Popen([self.name, "--version"], stdout=PIPE)
                ret, err = p.communicate()
                if (p.returncode == 0):
                    self.message = "Mermaid " + ret.decode()
            except OSError:
                pass
        super(Mermaid, self).greet()

    def evaluate(self, code):
        code_path = self.get_tmp_dir() + self.name + ".mmd"
        with open(code_path, 'w') as code_file:
            code_file.write(code)

        puppeteer_config = self.get_tmp_dir() + "puppeteer-config.json"
        with open(puppeteer_config, "w") as config_file:
            config_file.write('{"args": ["--no-sandbox"]}')

        png_path = self.get_png_path()
        if os.path.isfile (png_path):
            os.remove (png_path)
        cmd = [self.name, "-p", puppeteer_config, "-i", code_path, "-o", png_path]
        p = Popen(cmd, stderr=PIPE)
        out, err = p.communicate()
        if (p.returncode == 0):
            flush_file (self.get_png())
        else:
            flush_verbatim (err.decode())



#!/usr/bin/env python
###############################################################################
#
# MODULE      : tm_python.py
# DESCRIPTION : Initialize python plugin
# COPYRIGHT   : (C) 2004       Ero Carrera, ero@dkbza.org
#               (C) 2012       Adrian Soto
#               (C) 2014       Miguel de Benito Delgado, mdbenito@texmacs.org
#               (C) 2018-2020  Darcy Shen
#
# This software falls under the GNU general public license version 3 or later.
# It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
# in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.

import os
import sys
import platform
from os.path import exists
tmpy_home_path = os.environ.get("TEXMACS_HOME_PATH") + "/plugins/tmpy"
if (exists (tmpy_home_path)):
    sys.path.append(os.environ.get("TEXMACS_HOME_PATH") + "/plugins/")
else:
    sys.path.append(os.environ.get("TEXMACS_PATH") + "/plugins/")



import traceback
import string
import ast
from inspect   import ismodule, getsource, getsourcefile
from tmpy.compat import py_ver
from tmpy.capture import CaptureStdout
from tmpy.postscript import ps_out, PSOutDummy, pdf_out, FileOutDummy
from tmpy.completion import parse_complete_command, complete
from tmpy.protocol   import *

if py_ver == 2:
    flush_err ("Python 2 is no longer supported, please use Python 3")
    exit (-1)

# import logging as log
# log.basicConfig(filename='/tmp/tm_python.log',level=log.INFO)

def flush_output (data):
    """Do some parsing on the output according to its type.
    
    Non printable characters in unicode strings are escaped
    and objects of type None are not printed (so that procedure calls,
    as opposed to function calls, don't produce any output)."""

    if (data is None):
        flush_verbatim ("")
        return

    if isinstance (data, PSOutDummy):
        flush_ps (data.content)
    elif isinstance (data, FileOutDummy):
        if (data.content is None):
            flush_verbatim ("")
        else:
            flush_file (data.content)
    else:
        flush_verbatim (str(data).strip())

def as_scm_string (text):
    return '"%s"' % text.replace('\\', '\\\\').replace('"', '\\"')


def compile_help (text):
    cmd = 'help(%s)' % text
    out = {"help" : "", "src": "", "file": ""}

    try:
        out["help"] = CaptureStdout.capture (cmd, my_globals, "tm_python");
    except Exception as e:
        out ["help"] = 'No help for "%s": %s' % (text, e)

    try:
        out["src"] = eval ('getsource(%s)' % text,
                           my_globals, {'getsource' : getsource})
    except Exception as e:
        out["src"] = 'No code available for "%s": %s' % (text, e)

    try:
        # Todo: strip docstring from code
        out["file"] = eval ('getsourcefile(%s)' % text,
                            my_globals, {'getsourcefile' : getsourcefile})
    except Exception as e:
        out["file"] = 'Unable to access the code for "%s": %s' % (text, e)

    return dict (map (lambda k_v: (k_v[0], as_scm_string (k_v[1])), out.items()))


def my_eval (code, p_globals):
    '''Execute a script and return the value of the last expression'''

    block = ast.parse(code, mode='exec')
    if len(block.body) > 1 and isinstance(block.body[-1], ast.Expr):
        last = ast.Expression(block.body.pop().value)
        exec(compile(block, '<string>', mode='exec'), p_globals)
        return eval(compile(last, '<string>', mode='eval'), p_globals)
    else:
        return eval(code, p_globals)

__version__ = '3.0'
__author__ = 'Ero Carrera, Adrian Soto, Miguel de Benito Delgado, Darcy Shen'
my_globals   = {}

# We insert into the session's namespace the 'ps_out' method.
my_globals['ps_out'] = ps_out
my_globals['pdf_out'] = pdf_out

# As well as some documentation.
my_globals['__doc__'] = """A Python plugin for TeXmacs.
Provides autocompletion and embedding of PostScript data into the document,
e.g from files or from matplotlib.pyplot.
A rudimentary help window is also implemented: type the name of an object
with a question mark at the end to use it."""

text = 'import builtins as __builtins__'
CaptureStdout.capture (text, my_globals, "tm_python")

sys.stdout = os.fdopen (sys.stdout.fileno(), 'w')

###############################################################################
# Session start
###############################################################################
if (exists (tmpy_home_path)):
    flush_verbatim ("WARNING: You are under develop mode using " + tmpy_home_path)
    flush_newline (2)
flush_verbatim (f"Python { platform.python_version() } [{ sys.executable }] \n" +
               "Python plugin for TeXmacs.\n" +
               "Please see the documentation in Help -> Plugins -> Python")
flush_prompt (">>> ")
while True:
    line= input ()
    if not line:
        continue
    if line[0] == DATA_COMMAND:
        sf = parse_complete_command (line[1:])
        if sf[0] == 'complete':
            flush_scheme (complete (sf[1], sf[2], my_globals))
        continue
    elif line.endswith('?') and not line.strip().startswith('#'):
        if len(line) > 1:
            out= compile_help (line[:-1])
            flush_command ('(tmpy-open-help %s %s %s)' %
                           (out["help"], out["src"], out["file"]))
        else:
            flush_verbatim ('Type a name before the "?" to see the help')
        continue
    else:
        lines= [line]
        while line != "<EOF>":
            line= input ()
            if line == '': 
                continue
            lines.append (line)
        text = '\n'.join (lines[:-1])
        try: # Is it an expression?
            result= my_eval (text, my_globals)
        except:
            result= CaptureStdout.capture (text, my_globals, "tm_python")
        flush_output (result)
